/* This a kernel-module to handle the Universal A/D-Converter by             */
/* Frank Steinberg                                                           */
/*                                                                           */
/* Copyright (C) 2003 Mathias Paulsen (matpaulsen@gmx.de).                   */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify      */
/* it under the terms of the GNU General Public License as published by      */
/* the Free Software Foundation; either version 2 of the License, or         */
/* (at your option) any later version.                                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software               */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
/*                                                                           */
 
#include <linux/kernel.h>   
#include <linux/module.h>


#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif


#include <linux/fs.h>       
                            
#include <linux/wrapper.h>  

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a)*65536+(b)*256+(c))
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)
#include <asm/uaccess.h>  // needed by put_user 
#endif

#include <linux/delay.h> // needed by mdelay
#include <asm-i386/io.h> // needed by writeb and readb

static int io = 0 ; // module parameter io_port
static int Input = 0 ; // Input_port
static int Output = 0 ; // Output port
#define CLKhigh  1 
#define CShigh  2 
#define Clow 0
#define BitW 16


/* Setup Time (CS low before first I/O-Clock) */
#define TSU  2
/* I/O-Clock transition time */
#define TTIO 1 
/* duration of CS-High while conversion */
#define TWH 20


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
MODULE_PARM( io , "i" ); // macro to handle parameters
#endif

#define SUCCESS 0


#define DEVICE_NAME "ad_tlc549" // then name for our device

static int Device_Open = 0 ; // we don't want to open the device twice

// handles a device_open

static int device_open(struct inode *inode,
                       struct file *file)
{

  if (Device_Open)
    return -EBUSY ; // the device is still open

  Device_Open++ ; // increments the number of open devices

  MOD_INC_USE_COUNT ; // Macro: prevents unloading a module still in use

  return SUCCESS;
}

// handles a devie_close
// the number of parameters changed in newer kernels
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static int device_release(struct inode *inode,
                          struct file *file)
#else
static void device_release(struct inode *inode,
                           struct file *file)
#endif
{

  Device_Open -- ; // increments the number of open devices

  MOD_DEC_USE_COUNT ; // now we are able to unload the module

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
  return 0 ; // see declaration
#endif
}



#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static ssize_t device_read(struct file *file,
    char *buffer,    // The buffer to fill with data
    size_t length,   // The length of the buffer (not used)
    loff_t *offset)  // Our offset in the file 
#else
static int device_read(struct inode *inode,
                       struct file *file,
    char *buffer,   // The buffer to fill with data
                    
    int length)     // The length of the buffer
                     
#endif
{

  int bytes_read = 0 ; // we have to return the number of Bytes read

  int EV = 0 ;   // endvalue
  char OutBuf[80] ; // we copy a string to the user_space
  char *OutBufPtr ; // a pointer to the string
  int i ; // as usual
  int BV = 128 ; // bitvalue 

  // now we communicate with the tcl549 
 
  outb( CShigh , Output ) ; 
  udelay( TWH ) ; // min=17us (t_wH) 
  outb( Clow , Output ) ;
  
  udelay( TSU ) ;  // (t_SU) 

  for( i = 0 ; i < 8 ; ++i ) { 

  if( inb( Input ) & BitW ) EV += BV ;
  BV /= 2 ;
  outb( CLKhigh , Output ) ;
  udelay( TTIO ) ;  // t_t(I/O)
  outb( Clow , Output ) ;
  udelay( TTIO ) ;
  
  }
  

  OutBufPtr = OutBuf ; // puts the pointer to the first character
  sprintf( OutBuf , "%d\n" , EV ) ; // creates the the string 
  

// copies the string to the user_space
  while( *OutBufPtr && length ) { 
  
    put_user( *(OutBufPtr++)  , buffer++ ) ; 
    bytes_read++ ;
    length-- ;
  
  }

  return bytes_read ; // returns the number of bytes read - otherwise nothing happens

}




// module declarations

static int Major;
struct file_operations Fops = {

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
  NULL,   // owner
#endif

  NULL,   // seek

 device_read,
  NULL, //  device_write,
  NULL,   // readdir
  NULL,   // select
  NULL,   // ioctl
  NULL,   // mmap
  device_open,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
  NULL,   // flush
#endif

  device_release  // device_close

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
  ,
  NULL, // fsynk
  NULL, // fasync
  NULL, // lock
  NULL  // readv

#endif

};


// initialize the module
int init_module()
{


  char *PortName = "unknown" ;

  printk( "%s: Device driver for the universal A/D-converter by Frank Steiner.\n%s: Copyright by Mathias Paulsen (matpaulsen@gmx.de).\n" , DEVICE_NAME , DEVICE_NAME ) ; 
  printk( "%s: There is no warranty !!!\n" , DEVICE_NAME ) ;

  if( io == 0 ){


    printk( "%s: Too few parameters ...\n" , DEVICE_NAME ) ;
    printk( "%s: Try insmod %s io=0x??? next time.\n" , DEVICE_NAME , DEVICE_NAME ) ; 
    return -EBUSY ;
    
  } else {   // else io==NULL

    switch( io ){
      
    case 0x3f8: 
      PortName = "ttyS0" ;
      Input = io + 6 ;
      Output = io + 4 ;
      break ;
      
    case 0x2f8: 
      PortName = "ttyS1" ;
      Input = io + 6 ;
      Output = io + 4 ;
      break ;
      
    case 0x3e8:
      PortName = "ttyS2" ;
      Input = io + 6 ;
      Output = io + 4 ;
      break ;
      
    case 0x2e8:
      PortName = "ttyS3" ;
      Input = io + 6 ;
      Output = io + 4 ;
      break ;
      
    case 0x378:
      PortName = "lp0" ;
      Input = io + 1 ;
      Output = io ;
      break ;
      
    case 0x3bc:
      PortName = "lp0" ;
      Input = io + 1 ;
      Output = io ;
      break ;
      
    case 0x278:
      PortName = "lp1" ;
      Input = io + 1 ;
      Output = io ;
      break ;
      
    default:
      printk( "%s: Port %s (%x). Try another I/O-Address.\n" , DEVICE_NAME , PortName , io ) ;
      return -EBUSY ;
	}

    printk( "%s: using io=0x%x a.k.a. %s.\n" , DEVICE_NAME , io , PortName ) ;

    Input = io + 6 ;
    Output = io + 4 ;
   

    outb( 64 , io + 3 ) ; // turns TCL549 on

    /* Register the character device (atleast try) */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
    
    Major = register_chrdev(0,
                                 DEVICE_NAME,
                                 &Fops);

#else

  Major = module_register_chrdev(0,
                                 DEVICE_NAME,
                                 &Fops);
#endif
  
  if (Major < 0) {
    printk ("%s: %s %s device failed with %d\n", DEVICE_NAME , 
            "Sorry, registering the", DEVICE_NAME ,
            Major);
    return Major;
  }

  
  printk("%s: Enter\n%s: \"mknod /dev/%s c %d 0\"\n%s: to create a device file.\n" , DEVICE_NAME , DEVICE_NAME , DEVICE_NAME , Major , DEVICE_NAME ) ;
  
 

  } // end else io==0
  return 0;
}

// cleanup module
void cleanup_module()
{
  int ret;

  // unregister the device

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)

  ret = unregister_chrdev(Major, DEVICE_NAME);

#else

  ret = module_unregister_chrdev(Major, DEVICE_NAME);

#endif

  // If any error
  if (ret < 0)
    printk("%s: Error in unregister %s: %d\n", DEVICE_NAME ,  DEVICE_NAME , ret);

  outb( 0 , io + 3 ) ; // turns TLC549 off

}
